/*
  Name:  int DIFmyHealPCone

  Description:
   Calculates full and partial pixel intersected by a cone in the HEALPix 
   RING or NESTED sheme.
   This is a DIF custom version.

  Parameters:
   (i) DIF_Region &p:    Pointer to the DIF_Region class
   (i) float64 ra:     Cone center Right Ascension (degrees)
   (i) float64 dec:    Cone center Declination (degrees)
   (i) float64 radius: Cone radius (arcmin)

  Note:
   Npix = 12 * Nside^2
   Nside = 2^k, k = [0, ..., 29] -> order = WMAP resolution parameter
   Sky resolution: Omega=PI/(3Nside^2)
   For Nside=64 it is ~1 deg (12 * 64^2 = 49152 pixels).

   The minimum radius is TBD. At the moment it is ~ 2 cent-arcsec but 
   approximation problems in Healpix_Base2::query_disc suggest a larger
   value. To be fixed.

   Returned list of partial pixels could be > real ones, that of full pixels
   could be < real ones.
   If k is not in the allowed range then return empty lists
   (and not 0 code).

  Return 0 on success.


  LN@IASF-INAF, March 2009                        Last change: 27/03/2009
*/

#include <algorithm>


/* degrees to radians */
static const double DEG2RAD = 1.74532925199432957692369E-2;

/* Minimum cone radius (rad): ~ 2 cent-arcsec - this is TBC */
//static const double MIN_CONE_RAD = 5e-8;
static const double MIN_CONE_RAD = 1e-7;

#include "arr.h"
#include "geom_utils.h"
#include "healpix_base2.h"

using namespace std;

#include "dif.hh"


extern void difflist_i(vector<long long int>& v1,
                       vector<long long int>& v2,
                       vector<long long int>& vdiff);


int DIFmyHealPCone(DIF_Region &p)
{
  int64 my_nside;
  vector<int64> int_all;
  vector<long long int> all;
  vector<int64> int_flist;
  unsigned long int j;

  double ra = p.ra1;
  double dec = p.de1;
  double radius = p.rad;

// No available order: return
  if (p.params.size() == 0)
    return -1;

// Default is RING scheme
  int nested = 0, ik;


  if (p.getSchema() == DIF_HEALP_RING) { nested = 1; } // Nested?
  int k = p.params[0];
  ik = k;

//// Nested?
//  if (k < 0) {
//    nested = 1;
//    ik = -k;
//  } else
//    ik = k;

//cout << "Nested: "<< nested << "  Order: " << ik << endl;
  pointing ptg;

// Out of range: return here
  if ( ik > 29 )
    return -2;

  my_nside = 1 << ik;

  Healpix_Base2* base;

  if (nested)
//    Healpix_Base2 base (my_nside, NEST, SET_NSIDE);
    base = new Healpix_Base2(my_nside, NEST, SET_NSIDE);
  else
//    Healpix_Base2 base (my_nside, RING, SET_NSIDE);
    base = new Healpix_Base2(my_nside, RING, SET_NSIDE);

#ifdef DEBUG_PRINT
//  int order = base.nside2order(my_nside);
  cout << "Nside of map: " << my_nside << endl;
#endif

  ptg.theta = (90. - dec)*DEG2RAD;
  ptg.phi   = ra*DEG2RAD;
  double rad = radius/60.*DEG2RAD;

  if (rad < MIN_CONE_RAD) rad = MIN_CONE_RAD;

#ifdef DEBUG_PRINT
  if (rad == MIN_CONE_RAD)
    cout << "***Cone radius reset to " << MIN_CONE_RAD << " radians***" << endl;
#endif

// maximum angular distance between any pixel center and its corners
  double mpr = base->max_pixrad();

// All intersted pixels!
  base->query_disc(ptg, rad+mpr, int_all);

  for (j=0; j<int_all.size(); j++)
    all.push_back( int_all[j] );

// Sort array
  sort(all.begin(),all.end());


#ifdef DEBUG_PRINT
  cout << "ALL nodes:" << endl;
  for (j=0; j<all.size(); j++)
    cout << j << ": " << all[j] << endl;
#endif

// If just 1 then stop here
  if (all.size() == 0) {
#ifdef DEBUG_PRINT
    cout << "Error in query_disc_inclusive" << endl;
#endif
    return 1;
  }

// Decrease by the max pix. radius. This would give all pixel (approx ?!)
// fully covered by the disc.
//--  base.query_disc(ptg, rad-1.362*M_PI/(4*my_nside), int_flist);
  base->query_disc(ptg, rad-mpr, int_flist);

  vector<long long int> *flist, *plist;
  flist = &p.flist(k);
  plist = &p.plist(k);

  for (j=0; j<int_flist.size(); j++)
    flist->push_back( int_flist[j] );

// Sort array
  sort(flist->begin(),flist->end());
#ifdef DEBUG_PRINT
  cout << "Full nodes:" << endl;
  for (j=0; j<flist->size(); j++)
    cout << j << ": " << (*flist)[j] << endl;
#endif

// Difference -> partial nodes
  difflist_i(all,*flist,*plist);

#ifdef DEBUG_PRINT
  cout << "DIFmyHealPCone: N partial: " << plist->size() << endl;
  for (j=0; j<plist->size(); j++)
    cout << j << ": " << (*plist)[j] << endl;

  double m_area = M_PI / (3.*base->Nside()*base->Nside()) /
         DEG2RAD / DEG2RAD * 3600;
  double mpr_m = mpr / DEG2RAD * 60;
  double mpr_s = mpr_m * 60;
  cout << endl
       << "Total Nr of nodes: " << all.size() << "  Full: " << flist->size()
       << "  Partial: " << plist->size() << endl << endl
       << "Nside: " << base->Nside() << "  Npix: " << base->Npix() << endl
       << "Pixel area: " << m_area << " arcmin^2 ("
       << m_area * 3600 << " arcsec^2)" << endl
       << "Ang. res (SQRT(P_area)): " << sqrt(m_area) << " arcmin ("
       << sqrt(m_area)*60 << " arcsec)" << endl
       << "max_pixrad= " << mpr_m << " arcmin (" << mpr_s << " arcsec)" << endl;
#endif
 
  return 0;
}
