/*
  Name:  int getHealPBary, int getHealPBary1

  Description:
   Return the HEALPix barycenter (center) coordinates given scheme
   (RING or NESTED), order and pixel ID.
   getHealPBary1 always creates a new Healpix_Base2 base and destroys
   it on exit.

  Parameters:
 ( (i) char*& saved: if not NULL then re-use existing Healpix_Base2 )
   (i) int nested:       Scheme ID; if 0 then RING else NESTED
   (i) int k:            Resolution level of map in the range [0, 29]
   (i) long long int id: Pixel ID

   (o) double *bc_ra: HEALPix pixel barycenter RA (degrees)
   (o) double *bc_dec: HEALPix pixel barycenter Dec (degrees)

  Note:
    Use Healpix_Base2.

  Return 0 on success.


  LN@IASF-INAF, September 2008                      Last change: 25/03/2009
*/

#include "arr.h"
#include "healpix_base2.h"

/* degrees to radians */
static const double DEG2RAD = 1.74532925199432957692369E-2;

void cleanHealPUval(char*& saved);

using namespace std;


int getHealPBary(char*& saved, int nested, int k, long long int id,
                 double *bc_ra, double *bc_dec)
{
  long long int my_nside;
  Healpix_Base2* base;

  if (! saved) {

    if ((k < 0) || (k > 29))
      return -1;

    my_nside = 1 << k;

    if (nested) 
      base = new Healpix_Base2(my_nside, NEST, SET_NSIDE);
    else
      base = new Healpix_Base2(my_nside, RING, SET_NSIDE);

    saved = (char*) base;

  } else
    base = (Healpix_Base2*) saved;

  base->pix2ang_z_phi(id, *bc_dec, *bc_ra);
  *bc_ra /= DEG2RAD;
  *bc_dec = asin(*bc_dec)/DEG2RAD;
  
  return 0;
}


int getHealPBary1(int nested, int k, long long int id,
                  double *bc_ra, double *bc_dec)
{
  char* saved = NULL;
  int ret = getHealPBary(saved, nested, k, id, bc_ra, bc_dec);
  cleanHealPUval(saved);
  return ret;
}
