/*
   (double) skysep_h

  Purpose: gets the distance (in "arcmin") of 2 objects on a sphere.

  Usage: dist = skysep_h(theta1, phi1, theta2, phi2, coord_type)

  Note:
         If coord_type=0 then input values are in degrees, otherwise in radians.
         "theta" cannot be negative.
         All values are double but "coord_type" which is "short".

         Here the Haversine formula (R.W Sinnot, 1984) is used to get rid of
         singularities.


  LN@IASF-INAF, March 2007                    ( Last change: 11/10/2008 )
*/

#include <cmath>

/* 2*PI */
static const double PI2 = M_PI * 2.;
/* radians to degrees */
//#define RAD_DEG 57.2957795130823209

/* degrees to radians */
static const double DEG2RAD = 1.74532925199432957692369E-2;


double skysep_h(double theta1, double phi1, double theta2, double phi2,
                int coord_type)
{
  double a1r, b1r, a2r, b2r, radif, sin2a, sin2d;

  if (theta1 < 0. || theta2 < 0.) return -1.;

  if (! coord_type)
  {
    a1r = theta1*DEG2RAD;
    b1r = phi1*DEG2RAD;
    a2r = theta2*DEG2RAD;
    b2r = phi2*DEG2RAD;
  } else {
    a1r = theta1;
    b1r = phi1;
    a2r = theta2;
    b2r = phi2;
  }

  radif  = fabs(a2r-a1r);
  if (radif > M_PI) radif = PI2 - radif;
  sin2a = sin(radif/2.);
  sin2a *= sin2a;
  sin2d = sin((b2r-b1r)/2.);
  sin2d *= sin2d;
  return (2 * asin( sqrt(sin2d + cos(b1r)*cos(b2r)*sin2a) )/DEG2RAD * 60.);
}
